
/* Tiled window previews */
$tile_corner_radius: $base_border_radius + 1px;
.tile-preview {
  background-color: transparentize($selected_bg_color,0.5);
  border: 1px solid $selected_bg_color;
}

.tile-preview-left.on-primary {
  border-radius: $tile_corner_radius 0 0 0;
}

.tile-preview-right.on-primary {
  border-radius: 0 $tile_corner_radius 0 0;
}

.tile-preview-left.tile-preview-right.on-primary {
  border-radius: $tile_corner_radius $tile_corner_radius 0 0;
}